{% if bound_field.field.widget.input_type|default:None == "checkbox" %}
    {% include "otree/forms/rows/bootstrap_checkbox.html" with field=bound_field %}
{% else %}
    {% include "otree/forms/rows/bootstrap.html" with field=bound_field %}
{% endif %}
