{% load i18n %}
<p>
    <!-- .otree-btn-next is the CSS/jQuery selector for this button -->
    {% comment %}Translators: The text on the button the user clicks to get to the next page{% endcomment %}
    <button class="otree-btn-next btn btn-primary">{% blocktrans trimmed %}Next{% endblocktrans %}</button>
</p>
